/*   FILE: dmapd.c -- A DMAP server
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 01 January 2009 
 *
 * Copyright (c) 2009 W. Michael Petullo <new@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <libdmapsharing/dmap.h>

#include "dmapd-dmap-container-record.h"
#include "dmapd-dmap-container-db.h"
#include "dmapd-dpap-record.h"
#include "dmapd-dpap-record-factory.h"
#include "dmapd-daap-record.h"
#include "dmapd-daap-record-factory.h"
#include "dmapd-module.h"
#include "db-builder.h"
#include "av-meta-reader.h"
#include "photo-meta-reader.h"
#include "util.h"

#define CONFFILE SYSCONFDIR "/dmapd.conf"

/* For use when deciding whether to serve DAAP or DPAP. */
enum {
	DAAP,
	DPAP
};

GMainLoop *loop;

static GSList *music_dirs        = NULL;
static GSList *picture_dirs      = NULL;
static gchar *db_dir		 = "/var/db/dmapd";
static gchar *lockpath           = NULL;
static gchar *pidpath            = NULL;
static gchar *user               = NULL;
static gboolean foreground       = 0;
static gchar *share_name         = NULL;
static gchar *transcode_mimetype = NULL;
static gchar *av_module          = "gst";
static gchar *photo_module       = "vips";

pid_t child[2] = { 0, 0 };

static void
free_globals (void)
{
	g_debug ("Free'ing globals");

	slist_deep_free (music_dirs);
	slist_deep_free (picture_dirs);
}

static gboolean
add_dir (const gchar *option_name,
	 const gchar *value,
	 gpointer data,
	 GError **error)
{
	if (strcmp (option_name, "-m") == 0) {
		music_dirs = g_slist_append (music_dirs, g_strdup (value));
	} else if (strcmp (option_name, "-p") == 0) {
		picture_dirs = g_slist_append (picture_dirs, g_strdup (value));
	} else {
		g_error ("Unsupported directory option: %s.", option_name);
	}
	return TRUE;
}

/* FIXME: how to enumerate available transcoding formats? */
static GOptionEntry entries[] = {
	{ "foreground", 'f', 0, G_OPTION_ARG_NONE, &foreground, "Do not fork; remain in foreground", NULL },
	{ "name", 'n', 0, G_OPTION_ARG_STRING, &share_name, "Name of media shares", NULL },
	{ "music-dir", 'm', 0, G_OPTION_ARG_CALLBACK, add_dir, "Music directory", NULL },
	{ "picture-dir", 'p', 0, G_OPTION_ARG_CALLBACK, add_dir, "Picture directory", NULL },
	{ "lockpath", 'l', 0, G_OPTION_ARG_STRING, &lockpath, "Path to lockfile", NULL },
	{ "pidpath", 'i', 0, G_OPTION_ARG_STRING, &pidpath, "Path to PID file", NULL },
	{ "db-dir", 'd', 0, G_OPTION_ARG_STRING, &db_dir, "Media database directory", NULL },
	{ "user", 'u', 0, G_OPTION_ARG_STRING, &user, "User to run as", NULL },
	{ "transcode-mimetype", 't', 0, G_OPTION_ARG_STRING, &transcode_mimetype, "Target MIME type for real-time transcoding", NULL },
	{ NULL }
};

static char *
default_share_name ()
{
        const gchar *real_name;

        real_name = g_get_real_name ();
        if (strcmp (real_name, "Unknown") == 0) {
                real_name = g_get_user_name ();
        }

        return g_strdup_printf ("%s's Media", real_name);
}

static DMAPShare *
create_share (gint protocol, DMAPDb *db, DMAPContainerDb *container_db)
{
	char *name;

	DMAPShare *share = NULL;

	name = share_name ? share_name : default_share_name ();

	if (protocol == DAAP) {
		g_debug ("Initializing DAAP sharing (%d)", protocol);
		share = DMAP_SHARE (daap_share_new (name, NULL, db, container_db, transcode_mimetype));
	} else {
		g_debug ("Initializing DPAP sharing (%d)", protocol);
		share = DMAP_SHARE (dpap_share_new (name, NULL, db, container_db, transcode_mimetype));
	}

	g_free (name);

	return share;
}

static void
drop_root (const char *user) {
	struct passwd *pw;
	struct group *gr;

	if (!(pw = getpwnam(user))) {
		g_error ("Failed to find user %s", user);
	}

	if (!(gr = getgrnam("daemon"))) {	
		g_error ("Failed to find group %s", "daemon");
	}

	if (setgid (gr->gr_gid) < 0) {
		g_error ("Failed to set gid");
	}

	if (setuid (pw->pw_uid) < 0) {
		g_error ("Failed to set uid");
	}
}

static void
daemonize (void)
{
	int child, fd;
	char *_lockpath, *pid, *_pidpath;

	if (getppid () == 1)
		return; /* Already a daemon */

	child = fork ();

	if (child < 0)
		g_error ("Error forking");
	
	if (child > 0)
		exit (EXIT_SUCCESS);
	
	setsid ();

	chdir (RUNDIR);

	_lockpath = lockpath ? lockpath : LOCKPATH;
	fd = open (_lockpath, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	if (lockf (fd, F_TLOCK, 0) < 0)
		g_error ("Error opening lock file %s", _lockpath);
	
	_pidpath = pidpath ? pidpath : RUNDIR "/dmapd.pid";
	fd = open (_pidpath, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	pid = g_strdup_printf ("%d", getpid ());
	write (fd, pid, strlen (pid));
	g_free (pid);

	if (user != NULL) {
		drop_root (user);
	}
}

static pid_t
wait_on (pid_t pid, int protocol)
{
	int status;
	waitpid (pid, &status, 0);
	if (WIFEXITED (status)) {
		return WEXITSTATUS (status) ? 1 << protocol : EXIT_SUCCESS;
	} else {
		return EXIT_FAILURE;
	}
}

static void
debug_printf (const char *log_domain,
	      GLogLevelFlags log_level,
	      const gchar *message,
	      gpointer user_data)
{
	g_printf ("%s\n", message);
}

static void
debug_null (const char *log_domain,
	    GLogLevelFlags log_level,
	    const gchar *message,
	    gpointer user_data)
{
}

static GObject *
object_from_module (GType type, const gchar *fmt, ...)
{
	va_list ap;
	GType *filters;
	guint n_filters;
        gchar *db_module_name;
        gchar *db_module_path;
	DmapdModule *module;
	GObject *fnval;

	va_start (ap, fmt);

	db_module_name = g_strdup_vprintf (fmt, ap);
        db_module_path = g_module_build_path (MODULEDIR, db_module_name);

	module = dmapd_module_new (db_module_path);
	if (module == NULL || ! g_type_module_use (G_TYPE_MODULE (module))) {
                g_error ("Error opening %s", db_module_path);
	}

	filters = g_type_children (type, &n_filters);
	g_assert (n_filters == 1);
	fnval = g_object_new (filters[0], NULL);

        g_free (db_module_path);
        g_free (db_module_name);
	va_end (ap);

	return fnval;
}

static void
serve (gint protocol,
       DMAPRecordFactory *factory,
       GSList *media_dirs)
{
	GSList *l;
	DMAPDb *db;
	DMAPShare *share;
	DbBuilder *builder;
	DMAPContainerDb *container_db;

	db = DMAP_DB (dmapd_dmap_db_ghashtable_new (db_dir, factory));
	container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());
	builder = DB_BUILDER (object_from_module (TYPE_DB_BUILDER, "db-builder-%s", "gdir"));

	for (l = media_dirs; l; l = l->next) {
		db_builder_build_db_starting_at (builder, l->data, db, container_db, NULL);
	}

	loop = g_main_loop_new (NULL, FALSE);
	share = create_share (protocol, DMAP_DB (db), DMAP_CONTAINER_DB (container_db));

	g_main_loop_run (loop);

	g_object_unref (db);
	g_object_unref (container_db);
	g_object_unref (share);
	g_object_unref (builder);
}

static void
read_keyfile (void)
{
	GError *error = NULL;
	GKeyFile *keyfile;
	gchar **dir;
	gint i;
	gsize len;

	keyfile = g_key_file_new ();

	if (!g_key_file_load_from_file (keyfile, CONFFILE, G_KEY_FILE_NONE, &error)) {
		g_debug ("Could not read configuration file %s: %s", CONFFILE, error->message);
	} else {

		db_dir             = g_key_file_get_string (keyfile, "General", "Database-Dir", NULL);
		share_name         = g_key_file_get_string (keyfile, "General", "Share-Name", NULL);
		user               = g_key_file_get_string (keyfile, "General", "User", NULL);
		transcode_mimetype = g_key_file_get_string (keyfile, "Music", "Transcode-Mimetype", NULL);

		dir = g_key_file_get_string_list (keyfile, "Music", "Dirs", &len, NULL);
		for (i = 0; i < len; i++)
			add_dir ("-m", dir[i], NULL, &error);

		dir = g_key_file_get_string_list (keyfile, "Picture", "Dirs", &len, NULL);
		for (i = 0; i < len; i++)
			add_dir ("-p", dir[i], NULL, &error);
	}

	g_key_file_free (keyfile);
}

static void sigterm_handler_parent (int i)
{
	signal (SIGTERM, sigterm_handler_parent);

	if (child[DAAP]) {
		g_debug ("Killing PID %d", child[DAAP]);
		kill (child[DAAP], SIGTERM);
	}

	if (child[DPAP]) {
		g_debug ("Killing PID %d", child[DAAP]);
		kill (child[DPAP], SIGTERM);
	}
}

static void sigterm_handler_child (int i)
{
	signal (SIGTERM, sigterm_handler_parent);

	g_debug ("Received TERM signal");

	g_main_loop_quit (loop);
}

int main (int argc, char *argv[])
{
	int exitval = EXIT_SUCCESS;
	GError *error = NULL;
	GOptionContext *context;
	AVMetaReader *av_meta_reader = NULL;
	PhotoMetaReader *photo_meta_reader = NULL;

	g_type_init ();
	g_thread_init (NULL);

	if (getenv ("DMAPD_DEBUG") != NULL) {
		g_log_set_handler ("libdmapsharing", G_LOG_LEVEL_DEBUG, debug_printf, NULL);
		g_log_set_handler (NULL, G_LOG_LEVEL_DEBUG, debug_printf, NULL);
	} else {
		g_log_set_handler ("libdmapsharing", G_LOG_LEVEL_DEBUG, debug_null, NULL);
		g_log_set_handler (NULL, G_LOG_LEVEL_DEBUG, debug_null, NULL);
	}

	av_module = getenv ("DMAPD_AV_MODULE");
	av_module = av_module ? av_module : "gst";

	photo_module = getenv ("DMAPD_PHOTO_MODULE");
	photo_module = photo_module ? photo_module : "vips";

	context = g_option_context_new ("- serve media using DMAP");
	g_option_context_add_main_entries (context, entries, NULL);

	if (strcmp (av_module, "null") != 0) {
		GOptionGroup *group;
		av_meta_reader = AV_META_READER (object_from_module (TYPE_AV_META_READER, "av-meta-reader-%s", av_module));
		group = av_meta_reader_get_option_group (av_meta_reader);
		if (group)
			g_option_context_add_group (context, group);
	}

	if (strcmp (photo_module, "null") != 0) {
		GOptionGroup *group;
		photo_meta_reader = PHOTO_META_READER (object_from_module (TYPE_PHOTO_META_READER, "photo-meta-reader-%s", photo_module));
		group =  photo_meta_reader_get_option_group (photo_meta_reader);
		if (group)
			g_option_context_add_group (context, group);
	}

	read_keyfile ();

	if (! g_option_context_parse (context, &argc, &argv, &error)) {
		g_error ("Option parsing failed: %s", error->message);
	}

	g_option_context_free (context);

	/* FIXME: need to get libdmapsharing to support simultaneous DAAP and DPAP
	 * in one process. 
i	 */
	if (foreground && music_dirs && picture_dirs) {
		g_error ("Can't serve both music and photographs with the foreground option. Choose -m or -p");
	}

	if (! foreground) {
		daemonize ();
	}

	signal (SIGTERM, sigterm_handler_parent);

	if (music_dirs) {
#if WITH_DAAP
		/* FIXME: need to get libdmapsharing to support simultaneous DAAP and DPAP
		 * in one process.
		 */
		if (av_meta_reader == NULL)
			g_error ("Music directory specified but AV metadata reader module is 'null'");
		if (foreground || (child[DAAP] = fork ()) == 0) {
			DMAPRecordFactory *factory;
			signal (SIGTERM, sigterm_handler_child);
			stringleton_init ();
			factory = DMAP_RECORD_FACTORY (
					g_object_new (
						TYPE_DMAPD_DAAP_RECORD_FACTORY,
						"meta-reader",
						av_meta_reader,
						NULL));
			// g_object_unref (av_meta_reader);
			// g_object_unref (photo_meta_reader);
			serve (DAAP, factory, music_dirs);
			free_globals ();
			stringleton_deinit ();
			g_debug ("DAAP done");
		}
#else
		g_error ("DAAP support not present");
#endif
	}

	if (picture_dirs) {
#ifdef WITH_DPAP
		/* FIXME: need to get libdmapsharing to support simultaneous DAAP and DPAP
		 * in one process.
		 */
		if (photo_meta_reader == NULL)
			g_error ("Photo directory specified but photo metadata reader module is 'null'");
		if (foreground || (child[DPAP] = fork ()) == 0) {
			DMAPRecordFactory *factory;
			signal (SIGTERM, sigterm_handler_child);
			stringleton_init ();
			factory = DMAP_RECORD_FACTORY (
					g_object_new (
						TYPE_DMAPD_DPAP_RECORD_FACTORY,
						"meta-reader",
						photo_meta_reader,
						NULL));
			// g_object_unref (av_meta_reader);
			// g_object_unref (photo_meta_reader);
			serve (DPAP, factory, picture_dirs);
			free_globals();
			stringleton_deinit ();
			g_debug ("DPAP done");
		}
#else
		g_error ("DPAP support not present");
#endif
	}
	if (! foreground) {
		/* Forked, need to unref this process's memory: */
		free_globals ();
		// g_object_unref (av_meta_reader);
		// g_object_unref (photo_meta_reader);
	}

	if (child[DAAP]) {
		exitval |= wait_on (child[DAAP], DAAP);
		child[DAAP] = 0;
	}

	if (child[DPAP]) {
		exitval |= wait_on (child[DPAP], DPAP);
		child[DPAP] = 0;
	}

	g_debug ("Parent Exiting");

	exit(exitval);
}
