/*   FILE: util.c -- utility functions
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 01 January 2009 
 *
 * Copyright (c) 2009 W. Michael Petullo <new@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <libdmapsharing/dmap.h>

#include "util.h"

static GHashTable *stringleton;

GByteArray *
blob_add_atomic (GByteArray *blob, const guint8 *ptr, const size_t size)
{
	return g_byte_array_append (blob, ptr, size);
}

GByteArray *
blob_add_string (GByteArray *blob, const gchar *str)
{
	return g_byte_array_append (blob,
				   (const guint8 *) str,
				    strlen (str) + 1);
}

void
slist_deep_free (GSList *list)
{
	g_slist_foreach (list, (GFunc)g_free, NULL);
	g_slist_free (list);
}

void stringleton_init (void)
{
	static gboolean initialized = FALSE;

	if (! initialized) {
		stringleton = g_hash_table_new_full (g_str_hash,
						     g_str_equal,
						     g_free,
						     NULL);
	}

	initialized = TRUE;
}

const gchar *stringleton_ref (const gchar *str)
{
	gpointer key;
	gpointer val;

	/* NOTE: insert will free passed str if the key already exists,
	 * not existing key in hash table.
	 */
	if (g_hash_table_lookup_extended (stringleton, str, &key, &val)) {
		str = (gchar *) key;
		g_hash_table_insert (stringleton,
				    (gpointer) g_strdup (str),
		                     val + 1);
	} else {
		val = NULL;
		str = g_strdup (str);
		g_hash_table_insert (stringleton,
				    (gpointer) str,
		                     val + 1);
	}

	g_debug ("stringleton ref %s to %u", str, GPOINTER_TO_UINT (val));

	return str;
}

void stringleton_unref (const gchar *str)
{
	guint count;

	if (str != NULL) {
		count = GPOINTER_TO_UINT (g_hash_table_lookup (stringleton,
							      (gpointer) str));

		g_debug ("stringleton unref %s to %u", str, count - 1);

		/* NOTE: insert will free passed str if the key already exists,
		 * not existing key in hash table.
		 */
		if (count > 1) {
			g_hash_table_insert (stringleton,
					    (gpointer) g_strdup (str),
					     GUINT_TO_POINTER (count - 1));
		} else if (count == 1) {
			g_hash_table_remove (stringleton, (gpointer) str);
		}
	}
}

void stringleton_deinit (void)
{
	g_hash_table_destroy (stringleton);
}
